"""
The :mod:`sklearn.experimental` module provides importable modules that enable
the use of experimental features or estimators.

The features and estimators that are experimental aren't subject to
deprecation cycles. Use them at your own risks!
"""
